import serial
import time
import socket
#import datetime

from generic_module import Module

class HistModule(Module):
  def __init__(self, name, udp_ip, udp_port, serial_port, baud, interval, root_path, relevant_for_listeners):
    Module.__init__(self, name, root_path)
    self.udp_ip = udp_ip
    self.udp_port = udp_port
    self.serial_port = serial_port
    self.baud = baud
    self.interval = interval
    
    self.relevant_for_listeners = relevant_for_listeners # u, 0
  
  def histcmd(self, ser, cmnd, resp = '', to = 0.5):
    if ser.timeout != to:
      ser.timeout = to
    ser.flushInput()
    ser.write(cmnd + '\n')
    r = []
    s = ""
    while True:
      l = ser.readline().rstrip();
      if l == '': # EOF
        return r, s
      r.append(l)
      s = s + l
      if len(resp) > 0 and l.find(resp) != -1: # resp found
        return r, s
      if len(r) > 10: # safeguard
        return r, s
  
  def dict2confstr(self, section, d):
    r = '[' + section + ']' + '\n'
    for k,v in d.iteritems():
      r += k + '=' + v + '\n'
    return r
  
  def run(self):
    self.alive = True
    self.end = False
    even_odd_flag = 1
    try:
      ser = serial.Serial(self.serial_port, self.baud)
      r, s = self.histcmd(ser, 'PWRCHGLOG')
      #print s
      try:
        while True:
          #a temporary hack!
          if even_odd_flag == 0:
            even_odd_flag = 1
          else:
            even_odd_flag = 0
          if self.end:
            ser.close()
            self.alive = False
            return

          d = {}
          d['ts'] = time.strftime('%Y-%m-%d-%H-%M-%S')

          r, s = self.histcmd(ser, 'STAT')
          for p in r:
            parts = p.split("=")
            if parts[0] == "Ub__V" and "u" in self.relevant_for_listeners.lower() or parts[0] == "sw_ver":
              try:
                d[parts[0]] = parts[1]
              except Exception:
                pass
          for key, mod in self.downstream_modules_dict.iteritems():
            try:
              mod.set_hist_fw(d['sw_ver'])
            except Exception, ex:
              pass
            try:
              mod.set_voltage(d['Ub__V'])
            except Exception, ex:
              pass
            
          msg = self.dict2confstr('mydevice', d)
          sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
          try:
            sock.sendto(msg, (self.udp_ip, self.udp_port))
          finally:
            sock.close()
          self.zzzzz(self.interval)
      finally:
        ser.close()
    except Exception, ex:
      self.log(str(ex))
      self.alive = False

