import time
import socket
import datetime
import os
import json
import copy

from lxml import etree
from generic_module import Module

class ReceiveModule(Module):
  def __init__(self, name, typ, host, port, recv_buffer_size, save_original, root_path, offsets_dict, validity_tag):
    Module.__init__(self, name, root_path)
    self.typ = typ
    self.host = host
    self.port = port
    self.recv_buffer_size = recv_buffer_size
    self.save_original = save_original
    self.offsets_dict = offsets_dict
    self.validity_tag = validity_tag
    
    self.s = None
    self.parser = etree.XMLParser(remove_blank_text=True)
    self.am_i_a_server = False

  #def calculate_distance(self, s1, ts1, ts2, whlbse1, prepend2, append1):  
    #delta_t = datetime_object_2 - datetime_object_1
    #d_t_seconds = delta_t.total_seconds()
    #s1_fl = float(s1)
    #return d_t_seconds * s1_fl - float(whlbse1) - float(prepend2) - float(append1) #delta time * speed of the first vehicle ~ distance between the two vehicles when the second vehicle is at MP
      
  #def md5(self, fname):
    #hash_md5 = hashlib.md5()
    #with open(fname, "rb") as f:
      #for chunk in iter(lambda: f.read(4096), b""):
        #hash_md5.update(chunk)
    #return hash_md5.hexdigest()
  
  def set_end(self):
    self.end = True
    if self.am_i_a_server: 
      s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)  
      s.connect(("127.0.0.1", self.port))
    else:
      try:
        self.s.shutdown(socket.SHUT_WR)
      except Exception:
        pass    
  
  def generate_path(self, mod):
    pth = self.sites_path
    pth = pth + "/" + self.sname + "/ext"
    if not os.path.exists(pth):
      os.mkdir(pth)
    pth = pth + "/" + mod
    if not os.path.exists(pth):
      os.mkdir(pth)
    pth = pth + "/" + self.typ
    if not os.path.exists(pth):
      os.mkdir(pth)
    pth = pth + "/" + self.name 
    if not os.path.exists(pth):
      os.mkdir(pth)
    return pth
           
  def save_to_file(self, data_type, str_vehicles):
    if self.sname == "":
      return
    first_line = "<swd><site><name>" + self.sname + "</name><vehicles>\n"
    last_line = "</vehicles></site></swd>"
    try:
      date = datetime.datetime.now().strftime('%Y-%m-%d')
      path = self.generate_path(data_type)
      fd = open(path + "/" + date + ".xml", "a+")
      #remove last line          
      fd.seek(0, os.SEEK_END)
      pos = fd.tell() - 1
      while pos > 0 and fd.read(1) != "\n":
        pos -= 1
        fd.seek(pos, os.SEEK_SET)
      if pos > 0:
        fd.seek(pos, os.SEEK_SET)
        fd.truncate()
      vhs = ""
      for str_vehicle in str_vehicles:
        vhs = vhs + str_vehicle + "\n\n"
      #if file is empty, add first line
      if pos == fd.tell() - 1:
        fd.write(first_line + vhs + last_line)
      else:
        fd.write(vhs + last_line)
      fd.close()
    except Exception, ex:
      try:
        fd.close()
      except Exception:
        pass
      self.log(str(ex))
      
  def connect_tcp(self):
    self.s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)    
    try:
      self.s.connect((self.host, self.port))  
      self.s.settimeout(2)
      return 0
    except Exception, ex:
      self.log("Connection (" + self.name + ") failed with the following msg: " + str(ex)) 
      return 1
    
  def bind_udp(self):
    self.s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)    
    try:
      self.s.bind(("", self.port))  
      #self.s.settimeout(2)
      return 0
    except Exception, ex:
      self.log("Binding UDP socket (" + self.name + ") failed with the following msg: " + str(ex)) 
      return 1
    
  def bind_tcp(self):
    self.s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)    
    try:
      self.s.bind(('', self.port))  
      #self.s.settimeout(2)
      return 0
    except Exception, ex:
      self.log("Binding TCP socket (" + self.name + ") failed with the following msg: " + str(ex)) 
      return 1
    
  def siwim_new_xml_receiver(self):
    if self.connect_tcp() == 1:
      self.log("Connection failed.")
      return
    try:
      for key, mod in self.downstream_modules_dict.iteritems():
        mod.tell_offset(self.name, self.offsets_dict)
      line = ""
      cnt = 0 #for exiting thread
      while True:
        if self.end:
          return
        received = ""
        try:
          received = self.s.recv(self.recv_buffer_size)
        except socket.timeout:
          continue
        except socket.error, e:
          self.log(str(e))
          return
        else:
          line = line + received
          line = line.replace('<?xml version="1.0" ?>', "")
          starting_tag_id = line.find("<swd ")
          closing_tag_id = line.find("</swd>")
          original_xml = None
          if starting_tag_id != -1 and closing_tag_id != -1:
            dt_str = line[starting_tag_id:closing_tag_id + 6]
            try:
              original_xml = etree.fromstring(dt_str, self.parser)
              #save everything that we didn't use for xml 
              line = line[(closing_tag_id + 6):len(line)]
              cnt = 0
            except Exception:
              self.log("Bogus content, clearing the line.")
              #if data between starting and closing swd tag is "corba", we need to clear the line
              line = ""
              continue  
          else:
            #if there was no starting tag inside line, we need to clear the line
            if starting_tag_id == -1:
              line = ""
            cnt+=1
            if cnt == 8:
              self.log("Exiting thread: " + self.name)
              return
            continue
          self.sname = list(original_xml.iter("name"))[0].text
          original_vehicles = list(original_xml.iter("vehicle"))
          str_vehicles = list()
          for original_vehicle in original_vehicles:
            if self.validity_tag != "" and original_vehicle.find("wim").find(self.validity_tag) == None:
              continue
            str_vehicles.append(etree.tostring(original_vehicle))
            for key, mod in self.downstream_modules_dict.iteritems():
              mod.set_site_name(self.sname)
              mod.add_vehicle(copy.deepcopy(original_vehicle), self.name)
          if self.save_original == 1:          
            self.save_to_file("wim", str_vehicles)
    except Exception, ex:
      self.log("Exiting thread: " + self.name + ", " + str(ex))
      return
  
  def comark_lwh_receiver(self): 
    if self.connect_tcp() == 1:
      self.log("Connection failed.")
      return
    try:
      for key, mod in self.downstream_modules_dict.iteritems():
        mod.tell_offset(self.name, self.offsets_dict)
        #mod.set_site_name(self.sname)
        #mod.add_vehicle(vehicle, self.name)
      line = ""
      cnt = 0 #for exiting thread
      while True:
        if self.end:
          return
        received = ""
        try:
          received = self.s.recv(self.recv_buffer_size)
        except socket.timeout:
          continue
        except socket.error, e:
          self.log(str(e))
          return
        else:
          line = line + received
          starting_tag_id = line.find("<sensor ")
          closing_tag_id = line.find("</sensor>")
          original_xml = None
          if starting_tag_id != -1 and closing_tag_id != -1:
            dt_str = line[starting_tag_id:closing_tag_id + 9]
            try:
              original_xml = etree.fromstring(dt_str, self.parser)
              #save everything that we didn't use for xml 
              line = line[(closing_tag_id + 9):len(line)]
              cnt = 0
            except Exception, ex:
              #if data between starting and closing swd tag is "corba", we need to clear the line
              self.log("Bogus content, clearing the line.")
              #print line
              line = ""
              continue  
          else:
            #if there was no starting tag inside line, we need to clear the line
            if starting_tag_id == -1:
              #self.log("No starting tag, clearing the line.")
              line = ""
            cnt+=1
            if cnt == 8:
              self.log("Exiting thread: " + self.name)
              return
            #if closing_tag_id == -1:
              #self.log("No closing tag. Continuoing.")
            continue
          #add a container element in case line contains more than one message
          messages = list(original_xml.iter("sensor"))    
          str_vehicles = list()
          for message in messages:
            vehicle = etree.Element("vehicle")
            vehicle_1 = etree.Element("vehicle")
            lwh = etree.Element("lwh")
            vehicle.append(lwh)
            body = list(message.iter("transit_end"))[0]
            attributes = body.attrib
            #can't trust this; disregard
            if attributes["direction"] == "N":
              continue
            if attributes["speed"] == "0":
              continue
            for key in attributes.keys():
              if key == "time_iso_ms":
                ts = datetime.datetime.strptime(attributes[key], '%Y-%m-%dT%H:%M:%S.%f').strftime('%Y-%m-%d-%H-%M-%S-%f')
                ts = ts[:-3]
                lwh_element = etree.Element("ts")
                lwh_element.text = ts
                lwh.append(lwh_element)
              elif key == "speed":
                lwh_element = etree.Element("v")
                #convert km/h to m/s
                lwh_element.text = str(float(attributes[key])/3.6)
                lwh.append(lwh_element)
              elif key == "occupancy":
                lwh_element = etree.Element(key)
                lwh_element.text = attributes[key]
                lwh.append(lwh_element)
                lwh_element1 = etree.Element("siwim_i_add_to_offset")
                lwh_element1.text = str(float(attributes[key]) / 1000.0)
                lwh.append(lwh_element1)
              else: 
                lwh_element = etree.Element(key)
                lwh_element.text = attributes[key]
                lwh.append(lwh_element)
              #vehicle = etree.Element("vehicle")
              vehicle.append(lwh)
              vehicle_1.append(body)
            for key, mod in self.downstream_modules_dict.iteritems():
              mod.set_site_name(self.sname)
              mod.add_vehicle(copy.deepcopy(vehicle), self.name)
            str_vehicles.append(etree.tostring(vehicle_1))
              #self.log(etree.tostring(vehicle_1))
          if self.save_original == 1 and len(str_vehicles) > 0:   
            self.save_to_file("lwh", str_vehicles)
    except Exception, ex:
      self.log(str(ex))
      return        
  
  def macq_lines_to_dict(self, lines_list):
    lines_dict = dict()
    for line in lines_list:
      line_parts=line.split("=")
      lines_dict[line_parts[0]] = line_parts[1]
    return lines_dict
  
  def datetime_from_utc_to_local(self, utc_datetime):
    now_timestamp = time.time()
    offset = datetime.datetime.fromtimestamp(now_timestamp) - datetime.datetime.utcfromtimestamp(now_timestamp)
    return utc_datetime + offset
  
  def macq_anpr_receiver(self): 
    #import dateutil
    if self.bind_udp() == 1:
      self.log("Connection failed.")
      return
    try:
      for key, mod in self.downstream_modules_dict.iteritems():
        mod.tell_offset(self.name, self.offsets_dict)
      while True:

        if self.end:
          return        
        data, addr = self.s.recvfrom(self.recv_buffer_size)
        data_dict = self.macq_lines_to_dict(data.splitlines())        
        #for now; disregard status messages
        if data_dict["type"] == "status":
          continue        

        ts = data_dict["T"][0:4] + "-" + data_dict["T"][4:6] + "-" + data_dict["T"][6:8] + "-" + data_dict["T"][8:10] + "-" + data_dict["T"][10:12] + "-" + data_dict["T"][12:14] + "-" + data_dict["T"][14:17] 

        utc = datetime.datetime.strptime(ts, "%Y-%m-%d-%H-%M-%S-%f")

        ts = self.datetime_from_utc_to_local(utc).strftime("%Y-%m-%d-%H-%M-%S-%f")[:-3]

        lp = data_dict["P"]
        hgp = data_dict["D"]
        trg_ip = data_dict["Fc"]
        anpr_ip = data_dict["F"]
        lane = data_dict["lane"]

        vehicle_node = etree.Element("vehicle")
        vehicle_1_node = etree.Element("vehicle")
        anpr_node = etree.Element("anpr")
        anpr_1_node = etree.Element("anpr")
        
        vehicle_node.append(anpr_node)
        vehicle_1_node.append(anpr_1_node)
        
        ts_node = etree.Element("ts")
        lp_node = etree.Element("lp")
        hgp_node = etree.Element("hgp")
        anpr_ip_node = etree.Element("anpr_ip")
        trg_ip_node = etree.Element("trg_ip")
        lane_node = etree.Element("lane")
        
        ts_node.text = ts
        lp_node.text = lp
        hgp_node.text = hgp
        anpr_ip_node.text = anpr_ip
        trg_ip_node.text = trg_ip
        lane_node.text = str(lane)
        
        anpr_node.append(ts_node)
        anpr_node.append(lp_node)
        anpr_node.append(hgp_node)
        anpr_node.append(anpr_ip_node)
        anpr_node.append(trg_ip_node)
        anpr_node.append(lane_node)
        #print etree.tostring(vehicle_node)
        for key in data_dict.iterkeys():
          node = etree.Element(key)
          node.text = data_dict[key]
          anpr_1_node.append(node)        
        for key, mod in self.downstream_modules_dict.iteritems():
          mod.set_site_name(self.sname)
          mod.add_vehicle(copy.deepcopy(vehicle_node), self.name)        
        str_vehicles = list()
        str_vehicles.append(etree.tostring(vehicle_1_node))
        
        if self.save_original == 1 and len(str_vehicles) > 0:   
          self.save_to_file("anpr", str_vehicles)
        
    except Exception, ex:
      self.log(str(ex))
      return
  
  def technix_anpr_receiver(self): 
    if self.connect_tcp() == 1:
      self.log("Connection failed.")
      return
    try:
      for key, mod in self.downstream_modules_dict.iteritems():
        mod.tell_offset(self.name, self.offsets_dict)
      line = ""
      cnt = 0 #for exiting thread
      while True:
        if self.end:
          return
        received = ""
        try:
          received = self.s.recv(self.recv_buffer_size)
        except socket.timeout:
          continue
        except socket.error, e:
          self.log(str(e))
          return
        else:
          line = line + received
          starting_tag_id = line.find("<TAMPxml ")
          closing_tag_id = line.find("</TAMPxml>")
          original_xml = None
          if starting_tag_id != -1 and closing_tag_id != -1:
            dt_str = line[starting_tag_id:closing_tag_id + 10]
            try:
              original_xml = etree.fromstring(dt_str, self.parser)
              #save everything that we didn't use for xml 
              line = line[(closing_tag_id + 10):len(line)]
              cnt = 0
            except Exception, ex:
              #if data between starting and closing swd tag is "corba", we need to clear the line
              self.log("Bogus content, clearing the line.")
              #print line
              line = ""
              continue  
          else:
            #if there was no starting tag inside line, we need to clear the line
            if starting_tag_id == -1:
              #self.log("No starting tag, clearing the line.")
              line = ""
            cnt+=1
            if cnt == 8:
              self.log("Exiting thread: " + self.name)
              return
            #if closing_tag_id == -1:
              #self.log("No closing tag. Continuoing.")
            continue
          #add a container element in case line contains more than one message
          messages = list(original_xml.iter("Message"))    
          str_vehicles = list()
          for message in messages:
            if message.attrib["Type"] == "KeepAlive":
              top_element = etree.Element("TAMPxml", LocationId = "TAMPxmlClient")
              ka_message = etree.Element("Message", Type = "KeepAlive")
              top_element.append(ka_message)
              self.s.send(etree.tostring(top_element))
            elif message.attrib["Type"] == "ANPRevent":
              vehicle = etree.Element("vehicle")
              vehicle_1 = etree.Element("vehicle")
              anpr = etree.Element("anpr")
              vehicle.append(anpr)
              body = list(message.iter("Body"))[0]
              attributes = body.attrib
              kn_num = ""
              un_num = ""
              for key in attributes.keys():
                if key == "TRIGGER_TIME":
                  first_split = attributes[key].split(" ")
                  date_split = first_split[0].split(".")
                  time_split_1 = first_split[1].split(".")
                  time_split_2 = time_split_1[0].split(":")
                  ts = date_split[2] + "-" + date_split[1] + "-" + date_split[0] + "-" + time_split_2[0] + "-" + time_split_2[1] + "-" + time_split_2[2] + "-" + time_split_1[1]
                  anpr_element = etree.Element("ts")
                  anpr_element.text = ts
                  anpr.append(anpr_element)
                  continue
                elif key == "KN_NUM":
                  kn_num = attributes[key]
                elif key == "UN_NUM":
                  un_num = attributes[key]
                elif key == "ANPR_TXT":
                  lp_el = etree.Element("lp")
                  lp_el.text = attributes[key]
                  anpr.append(lp_el)
                elif key == "ANPR_IMAGE_PATH":
                  anpr_ip_el = etree.Element("anpr_ip")
                  anpr_ip_el.text = attributes[key]
                  anpr.append(anpr_ip_el)
                elif key == "ADR_IMAGE_PATH":
                  adr_ip_el = etree.Element("adr_ip")
                  adr_ip_el.text = attributes[key]
                  anpr.append(adr_ip_el)
                elif key == "COMBINED_TRG_IMAGE_PATH":
                  trg_ip_el = etree.Element("trg_ip")
                  trg_ip_el.text = attributes[key]
                  anpr.append(trg_ip_el)
                elif key == "LANE_ID":
                  lane_el = etree.Element("lane")
                  #ugly hack! just take the last character, due to a possibility of lane text actually being "Lane1"
                  #could be something else, but this should cover most cases
                  lane_el.text = attributes[key][len(attributes[key]) - 1]
                  anpr.append(lane_el)
              hdp_el = etree.Element("hgp")
              hdp_el.text = kn_num + " " + un_num
              anpr.append(hdp_el)
              #vehicle = etree.Element("vehicle")
              vehicle.append(anpr)
              vehicle_1.append(body)
              for key, mod in self.downstream_modules_dict.iteritems():
                mod.set_site_name(self.sname)
                mod.add_vehicle(copy.deepcopy(vehicle), self.name)
              str_vehicles.append(etree.tostring(vehicle_1))
          if self.save_original == 1 and len(str_vehicles) > 0:   
            self.save_to_file("anpr", str_vehicles)
    except Exception, ex:
      self.log(str(ex))
      return
    
  def cammra_anpr_receiver(self):
    self.am_i_a_server = True
    if self.bind_tcp() == 1:
      self.log("Connection failed.")
      return
    self.s.listen(3);
    for key, mod in self.downstream_modules_dict.iteritems():
      mod.tell_offset(self.name, self.offsets_dict)
    while True:
      conn, addr = self.s.accept()
      if self.end:
        try:
          conn.close()
          #self.s.close()
        except Exception:
          pass
        return
      data = ""
      while True:
        if self.end:
          return
        if len(data) > 2 and data.find("{\"") == -1:
          data = ""
        data+=conn.recv(self.recv_buffer_size)
        if "{\"" in data and "\n}" in data and data.find("{\"") < data.find("\n}"):
          relevant = data[data.find("{\""):data.find("\n}")+2]
          data_dict = json.loads(relevant)
          vehicle_node = etree.Element("vehicle")
          vehicle_1_node = etree.Element("vehicle")
          anpr_node = etree.Element("anpr")
          anpr_1_node = etree.Element("anpr")
          vehicle_node.append(anpr_node)
          vehicle_1_node.append(anpr_1_node)
          ts_node = etree.Element("ts")
          lp_node = etree.Element("lp")
          lane_node = etree.Element("lane")
          lane_node.text = data_dict["roiID"]
          lp_node.text = data_dict["plateUTF8"]
          ts = data_dict["datetime"]
          ts = ts[0:4] + "-" + ts[4:6] + "-" + ts[6:8] + "-" + ts[9:11] + "-" + ts[11:13] + "-" + ts[13:15] + "-" + ts[15:18]
          
          ts_node.text = ts
          
          anpr_node.append(ts_node)
          anpr_node.append(lp_node)
          anpr_node.append(lane_node)
          for key, val in data_dict.iteritems():
            node = etree.Element(key) 
            try:
              node.text = str(val)
            except Exception:
              continue
            anpr_1_node.append(node)
            #except Exception:
              #break
           
          str_vehicles = list()
          str_vehicles.append(etree.tostring(vehicle_1_node)) 
          #print str_vehicles
          for key, mod in self.downstream_modules_dict.iteritems():
            mod.set_site_name(self.sname)
            mod.add_vehicle(copy.deepcopy(vehicle_node), self.name)
          
          if self.save_original == 1 and len(str_vehicles) > 0:   
            self.save_to_file("anpr", str_vehicles)
          
          data = data[0:data.find("{\"")] + data[data.find("\n}"):-1]
          try:
            conn.close()
          except Exception:
            pass
          break
        
  def vaxtor_anpr_receiver(self):
    self.am_i_a_server = True
    #print "lel"
    if self.bind_tcp() == 1:
      self.log("Connection failed.")
      return
    self.s.listen(3);
    for key, mod in self.downstream_modules_dict.iteritems():
      mod.tell_offset(self.name, self.offsets_dict)
    while True:
      conn, addr = self.s.accept()
      if self.end:
        try:
          conn.close()
          #self.s.close()
        except Exception:
          pass
        return
      data = ""
      while True:
        if self.end:
          return
        if len(data) > 2 and data.find("{") == -1:
          data = ""
        data+=conn.recv(self.recv_buffer_size)
        if "{" in data and "}" in data and data.find("{") < data.find("}"):
          #print data
          relevant = data[data.find("{"):data.find("}")+1]
          relevant_parts = relevant[1:-1].split(",")
          ts = relevant_parts[0]
          lp = relevant_parts[1]
          lane = relevant_parts[2]
          country = relevant_parts[3].lower()
          
          npr_type = "alpr"
          if country == "adr":
            npr_type = "adr"
          #data_dict = json.loads(relevant)
          vehicle_node = etree.Element("vehicle")
          #vehicle_1_node = etree.Element("vehicle")
          anpr_node = etree.Element(npr_type)
          
          lp_node = None
          if npr_type == "adr":
            lp_node = etree.Element("hgp")
          else:
            lp_node = etree.Element("lp")
          #anpr_1_node = etree.Element("anpr")
          vehicle_node.append(anpr_node)
          #vehicle_1_node.append(anpr_1_node)

          if lp == "":
            lp = "/"
          
          ts_node = etree.Element("ts")
          lane_node = etree.Element("lane")

          ts = ts[0:4] + "-" + ts[5:7] + "-" + ts[8:10] + "-" + ts[11:13] + "-" + ts[14:16] + "-" + ts[17:19] + "-" + ts[20:23]
          ts_node.text = ts
          lp_node.text = lp
          lane_node.text = lane
          #country_node.text = country
          
          anpr_node.append(ts_node)
          anpr_node.append(lp_node)
          anpr_node.append(lane_node)
          
          str_vehicles = list()
          str_vehicles.append(etree.tostring(vehicle_node)) 
          #print str_vehicles
          #ignore for now
          for key, mod in self.downstream_modules_dict.iteritems():
            mod.set_site_name(self.sname)
            mod.add_vehicle(copy.deepcopy(vehicle_node), self.name + "_" + npr_type)
          
          if self.save_original == 1 and len(str_vehicles) > 0:   
            self.save_to_file("anpr", str_vehicles)

          data = data[0:data.find("{")] + data[data.find("}"):-1]
          try:
            conn.close()
          except Exception:
            pass
          break
     
  def run(self):
    self.alive = True
    self.end = False
    for key, mod in self.downstream_modules_dict.iteritems():
      mod.register_alarm(self.name)
    if self.typ == "technix":
      self.technix_anpr_receiver()
    elif self.typ == "cestel":
      self.siwim_new_xml_receiver()
    elif self.typ == "macq":
      self.macq_anpr_receiver()
    elif self.typ == "comark":
      self.comark_lwh_receiver()
    elif self.typ == "ffgroup":
      self.cammra_anpr_receiver()
    elif self.typ == "vaxtor":
      self.vaxtor_anpr_receiver()
    self.alive = False
    try:
      self.s.close()
    except Exception, ex:
      self.log(str(ex))
      pass
